----------
--- This work by Evan Leybourn is licensed under a Creative Commons Attribution-ShareAlike 3.0 Australia License 
--- <http://creativecommons.org/licenses/by-sa/3.0/au/>
----------


----------
-- DROP TABLES
--
-- Drop any existing tables. This is used to ensure the core tables
-- in the demo database are reset. 
----------
DROP TABLE Transactions;
DROP TABLE Accounts;
DROP TABLE Customers;
DROP TABLE WorksAt;
DROP TABLE Branches;
DROP TABLE Staff;

----------
-- CREATE TABLES
--
-- Create the demo tables. Note that these tables have deliberate 
-- logical errors
----------
CREATE TABLE Staff (
	staff_id INT IDENTITY(1,1) PRIMARY KEY,
	given_name NVARCHAR(50),
	family_name NVARCHAR(50),
	role NVARCHAR(50)
)

CREATE TABLE Branches (
	name VARCHAR(50),
	location VARCHAR(50),
	manager INT REFERENCES Staff
		ON DELETE SET NULL
		ON UPDATE CASCADE,
      CONSTRAINT branchpk PRIMARY KEY (name)
)

CREATE TABLE WorksAt (
	branch_id VARCHAR(50) REFERENCES Branches
		ON UPDATE CASCADE
		ON DELETE CASCADE,
	staff_id INT REFERENCES Staff
		ON UPDATE NO ACTION
		ON DELETE NO ACTION,
	CONSTRAINT worksatpk PRIMARY KEY (branch_id, staff_id)
)

CREATE TABLE Customers (
	customer_id INT IDENTITY(1,1),
	given_name VARCHAR(50),
	family_name VARCHAR(50),
	gender CHAR(1),
	dob DATE,
	age INT,
	branch_id VARCHAR(50) REFERENCES Branches
		ON UPDATE CASCADE
		ON DELETE CASCADE,
      CONSTRAINT customerpk PRIMARY KEY (customer_id)
)

CREATE TABLE Accounts (
	account_id BIGINT IDENTITY(1,1) PRIMARY KEY,
	customer_id INT REFERENCES Customers
		ON UPDATE NO ACTION
		ON DELETE NO ACTION,
	balance NUMERIC CHECK (balance > 0),
	opened_by INT REFERENCES Staff
		ON UPDATE NO ACTION
		ON DELETE NO ACTION
)

CREATE TABLE Transactions (
	account_id BIGINT REFERENCES Accounts
		ON UPDATE NO ACTION
		ON DELETE NO ACTION,
	transaction_amt NUMERIC,
	date DATETIME
) 

----------
-- LOAD DATA
--
-- Populate the database with demo data. 
----------
INSERT INTO Staff VALUES ('Isaac', 'Asmiov', 'Manager');
INSERT INTO Staff VALUES ('Ray', 'Bradbury', 'Teller');
INSERT INTO Staff VALUES ('Arthur', 'Clarke', 'Teller');
INSERT INTO Staff VALUES ('Samuel', 'Delany', 'Customer Service');
INSERT INTO Staff VALUES ('Greg', 'Egan', 'Teller');

INSERT INTO Branches VALUES ('Canberra City', 'Canberra', '1');
INSERT INTO Branches VALUES ('St Lucia', 'Brisbane', '2');
INSERT INTO Branches VALUES ('Haymarket', 'Sydney', null);
INSERT INTO Branches VALUES ('Coburg', 'Melbourne', null);

INSERT INTO WorksAt VALUES ('Canberra City', '1');
INSERT INTO WorksAt VALUES ('St Lucia', '2');
INSERT INTO WorksAt VALUES ('Haymarket', '3');
INSERT INTO WorksAt VALUES ('Coburg', '2');

INSERT INTO Customers VALUES ('Esther', 'Friesner', 'F', '01-01-1980', NULL, 'Canberra City');
INSERT INTO Customers VALUES ('William', 'Gibson', 'M', '01-01-1970', NULL, 'St Lucia');
INSERT INTO Customers VALUES ('Peter', 'Hamilton', 'M', '01-01-1969', NULL, 'St Lucia' );
INSERT INTO Customers VALUES ('Simon', 'Ings', 'M', '01-01-1971', NULL, NULL);
INSERT INTO Customers VALUES ('Gweneth', 'Jones', 'F', '01-01-1980', NULL, NULL);
INSERT INTO Customers VALUES ('Forest', 'Temmer', 'M', '01-01-1990', NULL, NULL);

INSERT INTO Accounts VALUES ('1', 0, '1');
INSERT INTO Accounts VALUES ('1', 0, '1');
INSERT INTO Accounts VALUES ('2', 0, '1');
INSERT INTO Accounts VALUES ('3', 0, '2');
INSERT INTO Accounts VALUES ('4', 0, '2');
INSERT INTO Accounts VALUES ('5', 0, '3');

INSERT INTO Transactions VALUES ('1', '100.00', getdate());
INSERT INTO Transactions VALUES ('1', '200.00', getdate());
INSERT INTO Transactions VALUES ('1', '-50.00', getdate());
INSERT INTO Transactions VALUES ('2', '100.00', getdate());
INSERT INTO Transactions VALUES ('2', '200.00', getdate());
INSERT INTO Transactions VALUES ('2', '200.00', getdate());
INSERT INTO Transactions VALUES ('2', '-500.00', getdate());
INSERT INTO Transactions VALUES ('2', '100.00', getdate());
INSERT INTO Transactions VALUES ('3', '100.00', getdate());
INSERT INTO Transactions VALUES ('3', '200.00', getdate());
INSERT INTO Transactions VALUES ('3', '200.00', getdate());
INSERT INTO Transactions VALUES ('3', '150.00', getdate());
INSERT INTO Transactions VALUES ('3', '-50.00', getdate());
INSERT INTO Transactions VALUES ('3', '-50.00', getdate());
INSERT INTO Transactions VALUES ('4', '100.00', getdate());
INSERT INTO Transactions VALUES ('4', '200.00', getdate());
INSERT INTO Transactions VALUES ('4', '200.00', getdate());
INSERT INTO Transactions VALUES ('4', '-250.00', getdate());
INSERT INTO Transactions VALUES ('4', '-50.00', getdate());
INSERT INTO Transactions VALUES ('5', '100.00', getdate());
INSERT INTO Transactions VALUES ('5', '200.00', getdate());
INSERT INTO Transactions VALUES ('5', '200.00', getdate());
INSERT INTO Transactions VALUES ('5', '250.00', getdate());
INSERT INTO Transactions VALUES ('5', '250.00', getdate());
INSERT INTO Transactions VALUES ('5', '200.00', getdate());
INSERT INTO Transactions VALUES ('1', '-2000.00', getdate());