----------
--- This work by Evan Leybourn is licensed under a Creative Commons Attribution-ShareAlike 3.0 Australia License 
--- <http://creativecommons.org/licenses/by-sa/3.0/au/>
----------


----------
-- INITIAL STEPS 
--
-- 1: Login to SSMS
-- 2: Create corebank database
-- 3: Open "New Query" window
-- 4: Run createbank.sql
-- 5: <optional> Attach AdventureWorks2012 database (ensure 
--    persmissions are correct and log file can be created)
----------

----------
-- NAVIGATE SSMS
--
-- 1: Create and view ER Diagram
-- 2: Navigate tables heirarchy
-- 3: Investigate table and column options
-- 4: Investigate table and column constraints (primary keys, foreign 
--    keys, check constraints)
-- 5: Navigate views heirarchy
----------

----------
-- INFORMATION SCHEMA
--
-- Explore the common information_schama views
----------
SELECT * FROM INFORMATION_SCHEMA.TABLES
SELECT * FROM INFORMATION_SCHEMA.COLUMNS
SELECT * FROM INFORMATION_SCHEMA.VIEWS
SELECT * FROM INFORMATION_SCHEMA.KEY_COLUMN_USAGE
SELECT * FROM INFORMATION_SCHEMA.CHECK_CONSTRAINTS

-- Explore the uncommon information_schama views
SELECT * FROM INFORMATION_SCHEMA.CONSTRAINT_COLUMN_USAGE
SELECT * FROM INFORMATION_SCHEMA.CONSTRAINT_TABLE_USAGE
SELECT * FROM INFORMATION_SCHEMA.REFERENTIAL_CONSTRAINTS
SELECT * FROM INFORMATION_SCHEMA.TABLE_CONSTRAINTS
SELECT * FROM INFORMATION_SCHEMA.TABLE_PRIVILEGES
SELECT * FROM INFORMATION_SCHEMA.VIEW_COLUMN_USAGE
SELECT * FROM INFORMATION_SCHEMA.VIEW_TABLE_USAGE

-- Explore the remaining information_schama views
SELECT * FROM INFORMATION_SCHEMA.COLUMN_DOMAIN_USAGE
SELECT * FROM INFORMATION_SCHEMA.DOMAIN_CONSTRAINTS
SELECT * FROM INFORMATION_SCHEMA.DOMAINS
SELECT * FROM INFORMATION_SCHEMA.COLUMN_PRIVILEGES
SELECT * FROM INFORMATION_SCHEMA.PARAMETERS
SELECT * FROM INFORMATION_SCHEMA.ROUTINE_COLUMNS
SELECT * FROM INFORMATION_SCHEMA.ROUTINES
SELECT * FROM INFORMATION_SCHEMA.SCHEMATA

--Q1: Where would you find a list of all check constraints
SELECT * FROM INFORMATION_SCHEMA.CHECK_CONSTRAINTS
--Q2: Where would you find a list of all columns
SELECT * FROM INFORMATION_SCHEMA.COLUMNS
--Q3: Where would you find details of foreign keys
SELECT * FROM INFORMATION_SCHEMA.KEY_COLUMN_USAGE
SELECT * FROM INFORMATION_SCHEMA.CONSTRAINT_COLUMN_USAGE
SELECT * FROM INFORMATION_SCHEMA.REFERENTIAL_CONSTRAINTS


----------
-- TRANSACTIONS
--
-- Show the syntax to wrap queries in transactions
----------
-- This will cause a deviation between the balance and the sum of all
-- transactions per account (check constraint on Acounts). 
INSERT INTO Transactions VALUES (1, -1000000, getdate())
UPDATE Accounts SET balance = balance - 100000 WHERE account_id=1;
-- Proof
SELECT * FROM transactions WHERE account_id=1
DELETE FROM transactions WHERE transaction_amt=-1000000
-- Solution: wrap the queries in a transaction
BEGIN TRANSACTION 
	BEGIN TRY
		INSERT INTO Transactions VALUES (1, -1000000, getdate())
		UPDATE Accounts SET balance = balance - 100000 WHERE account_id=1;
		COMMIT TRANSACTION 
	END TRY
	BEGIN CATCH
		ROLLBACK TRANSACTION 
	END CATCH
--Proof
SELECT * FROM transactions WHERE account_id=1
DELETE FROM transactions WHERE transaction_amt=1000000
-- And let's try it for a successful transaction
BEGIN TRANSACTION name
	BEGIN TRY
		INSERT INTO Transactions VALUES (1, 1000000, getdate())
		UPDATE Accounts SET balance = balance + 100000 WHERE account_id=1;
		COMMIT TRANSACTION name
	END TRY
	BEGIN CATCH
		ROLLBACK TRANSACTION name
	END CATCH
--Proof
SELECT * FROM transactions WHERE account_id=1


