----------
--- This work by Evan Leybourn is licensed under a Creative Commons Attribution-ShareAlike 3.0 Australia License 
--- <http://creativecommons.org/licenses/by-sa/3.0/au/>
----------


----------
-- SQL KEYWORDS
--
-- Show that SQL keywords are restricted
----------
CREATE TABLE select (
	id INT
)
CREATE TABLE foo (
	select INT
)
-- But if you really want to.
CREATE TABLE foo (
	"select" INT
)


----------
-- SQL DATATYPES
--
-- What would be a suitable type for?
-- Family name – nvarchar(50)
-- DOB - date
-- Age - int
-- Salary - numeric(8,2)
-- Timestamp - datetime
-- Gender field - char(1), boolean, varchar(20): depends on how you 
--                define gender
-- Role Description – nvarchar(max) or ntext
-- Student ID eg s2006756 – char(8)
-- Transaction ID – bigint
-- Active Account – boolean
-- Last Login – datetime
-- File attachement – varbinary(max) or varchar(100) and store the 
--                    file on the filesystem
----------


----------
-- SQL OPERATORS: MATHS
--
-- Show the common mathematical operators
----------
SELECT 1+1
SELECT sin(180), cos(180), tan(180)
SELECT abs(-100)
SELECT CEILING(10.55), FLOOR(10.55), ROUND(10.55, 0), ROUND(10.55, 1)
SELECT power(2, 3), SELECT SQRT(9)
SELECT random()
-- Q1: How would you return a random whole number between 1 and 100
SELECT round(random()*100, 0)


----------
-- SQL OPERATORS: STRINGS
--
-- Show the common string operators
----------
SELECT 'Hello' + ' World'
SELECT len('Hello')
SELECT SUBSTRING('Hello', 4, 2)
SELECT lower('NSW'), upper('nsw')
SELECT CHARINDEX(' ', 'Hello World')
SELECT LEFT('Hello World', 5), RIGHT('Hello World', 5)
-- Q1: How would you strip the first word off a string
SELECT LEFT('Hello World', CHARINDEX(' ', 'Hello World'))


----------
-- SQL OPERATORS: DATES
--
-- Show the common date operators
----------
SELECT getdate()
SELECT datediff(yyyy, '01-01-1970', '01-01-1999')
SELECT day(getdate()), month(getdate()), year(getdate()), datepart(yyyy, getdate()), datepart(dy, getdate())
--Q1: How would you get someones age from their dob
SELECT datediff(yyyy, '01-01-1970', getdate())


----------
-- AGGREGATION
--
-- Show the aggregation (and group by) syntax
----------
SELECT count(*), sum(transaction_amt), avg(transaction_amt), min(transaction_amt), max(transaction_amt) from transactions
SELECT sum(*), account_id 
	FROM Accounts
	GROUP BY account_id


----------
-- NULL
--
-- Null will always return NULL (expect in aggregation)
----------
SELECT 1+NULL, 3/NULL
-- Proof aggregation should return the same value
SELECT sum(transaction_amt) FROM transactions
INSERT INTO transactions VALUES (1, null, getdate())
SELECT * FROM transactions
SELECT sum(transaction_amt) FROM transactions


----------
-- UNICODE
--
-- How to insert unicode strings into the database. This should be 
-- the default unless there is a valid design reason not to.
----------
SELECT * FROM Staff
-- Unicode string
INSERT INTO Staff VALUES (N'???', N'???', 'Teller')
-- Proof
SELECT * FROM Staff
-- Unicode string, but not encapsulated correctly
INSERT INTO Staff VALUES ('???', '???', 'Teller')
SELECT * FROM Staff


----------
-- SQL DEMONSTRATION
--
-- Demonstrate all the SQL functions that have been discussed to 
-- date. Focus on JOINS. 
----------
SELECT * FROM Customers
SELECT * FROM Customers WHERE gender IN ('M', 'F');
SELECT * FROM Customers c, Branches b
SELECT * FROM Customers , Accounts a 
	WHERE c.customer_id = a.customer_id
SELECT * FROM Accounts a1, Accounts a2 
	WHERE a1.customer_id = a2.customer_id


----------
-- SQL EXERCISES
--
-- Test all the SQL functions that have been discussed to date. Focus 
-- on JOINS. 
----------
--Q1. Return all branches
SELECT * FROM Branches
--Q2. List of all Female Customers
SELECT * FROM Customers WHERE gender='F'
--Q3. List of all Male Customers over 40
SELECT * 
	FROM Customers 
	WHERE gender='M' AND datediff(yy, dob, getdate()) > 40
--Q4. List of all Customers sorted by given name
SELECT * FROM Customers ORDER BY given_name
--Q5. List of all Customers with a home branch
SELECT * FROM Customers 
	WHERE branch_id is not null ---- or
SELECT * FROM Customers c, Branches b 
	WHERE c.branch_id = b.name
--Q6. List of all Customers with their home branch (or not)
SELECT * 
	FROM Customers c LEFT OUTER JOIN Branches b ON c.branch_id = b.name
--Q7. Average Transaction size
SELECT avg(transaction_amt) FROM Transactions
--Q8. List of all Customers without a Branch
SELECT * FROM Customers WHERE branch_id is null 


----------
-- MORE SQL EXERCISES
--
-- It never ends 
----------
--Q1. A List of unique roles in the system
SELECT DISTINCT role from Staff
--Q2. List of all people (Customers & Staff) in the system
SELECT customer_id, family_name, given_name FROM Customers
	UNION
	SELECT staff_id, family_name, given_name FROM Staff
--Q2(ADVANCED): Include the type of person
SELECT customer_id, family_name, given_name, 'Customer' FROM Customers
	UNION
	SELECT staff_id, family_name, given_name, 'Staff' FROM Staff
--Q3. The list of all Customers without Accounts
SELECT * 
	FROM Customers 
	WHERE customer_id NOT IN (SELECT customer_id FROM Accounts)
--Q4. The correct balance per account
SELECT sum(transaction_amt), account_id 
	FROM Transactions 
	GROUP BY account_id
--Q5. The Customer with the highest balance (by transaction)
SELECT TOP 1 sum(t.transaction_amt) as sum, c.customer_id 
	FROM Transactions t, Accounts a, Customers c 
	WHERE t.account_id=a.account_id AND c.customer_id=a.customer_id
	GROUP BY c.customer_id
	ORDER BY sum desc
--Q6. Number of Staff per location
SELECT count(*), branch_id FROM worksat GROUP BY branch_id
--Q7. List of all Customers whose given name contains ‘e’
SELECT * FROM Customers WHERE given_name LIKE '%e%'
--Q8. List of all Customers whose name 2nd character is ‘e’
SELECT * FROM Customers WHERE given_name LIKE '_e%'
--Q9. List of all Customers whose name does start with ‘e’
SELECT * FROM Customers WHERE given_name LIKE '[^e]%'
--Q10. List of Customers with their highest balance Account and lowest balance Account
SELECT t.customer_id, min(t.sumt) as min, max(t.sumt) as max FROM
	(
	SELECT c.customer_id, a.account_id, sum(t.transaction_amt) as sumt
		FROM Customers c, Accounts a, Transactions t
		WHERE c.customer_id=a.customer_id AND a.account_id=t.account_id
		GROUP BY c.customer_id, a.account_id
	) as t
		GROUP BY customer_id


----------
-- QUERY EXECUTION PLAN
--
-- Show QEP
-- Show how to read plan (backwards)
-- Explain how Cost % works
-- Show the difference between table scan and index scan
----------


----------
-- BROKEN SQL
--
-- Identify the errors in the following statements
----------
--Q1: Missing GROUP BY
SELECT account_id, avg(transaction_amt) FROM Transactions
--Q2: IS NOT NULL not != NULL
SELECT * FROM Customers WHERE branch_id != null 
--Q3: DISTINCT not UNIQUE
SELECT UNIQUE role from Staff
--Q4: ' not "
SELECT customer_id, family_name, given_name, "Customer" FROM Customers
	UNION
	SELECT staff_id, family_name, given_name, "Staff" FROM Staff
--Q5: Incorrect comparison
SELECT * FROM Customers WHERE customer_id NOT IN (SELECT account_id FROM Accounts)
--Q6: Case insensitivity	
SELECT * 
	FROM Customers 
	WHERE gender='f'
--Q7: Missing JOIN fields
SELECT * 
	FROM Customers c, Accounts a 	
	WHERE c.customer_id = 1
--Q8: Incorrect join fields
SELECT * 
	FROM Customers c JOIN Transactions t ON (t.account_id=c.customer_id)
--Q9: Different number of fields between the UNION
SELECT * FROM Customers
UNION
SELECT * FROM Staff
--Q10: Undifferentiated tables
SELECT * 
	FROM Accounts, Accounts
	WHERE accounts.customer_id=accounts.customer_id
--Q11: yy not 'yy'
SELECT * FROM Customers WHERE gender='M' AND datediff('yy', dob, getdate()) > 40
--Q12: missing as t
SELECT t.customer_id, min(t.sumt) as min, max(t.sumt) as max FROM
	(
	SELECT c.customer_id, a.account_id, sum(t.transaction_amt) as sumt
		FROM Customers c, Accounts a, Transactions t
		WHERE c.customer_id=a.customer_id AND a.account_id=t.account_id
		GROUP BY c.customer_id, a.account_id
	) 
		GROUP BY customer_id
--Q13: Incorrect join 
SELECT t.customer_id, min(t.sumt) as min, max(t.sumt) as max FROM
	(
	SELECT c.customer_id, a.account_id, sum(t.transaction_amt) as sumt
		FROM Customers c, Accounts a, Transactions t
		WHERE c.customer_id=a.customer_id AND t.account_id=t.account_id
		GROUP BY c.customer_id, a.account_id
	) as t
		GROUP BY customer_id
--Q14: sum() not min()
SELECT t.customer_id, min(t.sumt) as min, max(t.sumt) as max FROM
	(
	SELECT c.customer_id, a.account_id, min(t.transaction_amt) as sumt
		FROM Customers c, Accounts a, Transactions t
		WHERE c.customer_id=a.customer_id AND a.account_id=t.account_id
		GROUP BY c.customer_id, a.account_id
	) as t
		GROUP BY customer_id
--Q15: Wrong database - Note: change the database to 'master'
SELECT * FROM Customers

----------
-- TEST CASE STRUCTURE
--
-- Some options on how to wrap SQL in test cases. 
----------
-- Type A: Returns result set
SELECT 
	CASE 
		WHEN <INCLUDE TEST CHECK <BASED ON TEST QUERY RESULTS> HERE>  THEN 'Test Pass'
		WHEN <INCLUDE TEST CHECK <BASED ON TEST QUERY RESULTS> HERE> THEN 'Test Fail'
	END
	FROM (
		<INCLUDE TEST QUERY HERE>
	) as t

-- Type B: Returns Exception (RAISERROR)
IF <INCLUDE TEST CHECK HERE>(<INCLUDE TEST QUERY HERE>)
	RAISERROR ('Test Fail', 11, 1)
ELSE
	PRINT 'Test Pass'

-- Type C: Returns Exception with customer message
DECLARE @msg varchar(255)
DECLARE @result int
SET @result = 0
SET @msg = 'Test Fail: '

SET @result = (SELECT count(*) FROM (
		<INCLUDE TEST QUERY HERE>
	) as t)

IF <INCLUDE TEST CHECK HERE>
	BEGIN
		SET @msg = @msg + cast(@result as varchar) + ' records found. Expected 0'
		RAISERROR (@msg , 11, 100)
	END
ELSE
	PRINT 'Test Pass'


----------
-- TEST CASE DEMO
--
-- Show how you can wrap simple SQL in a test case wrapper
----------
-- Setup (compare current DB schema to a previous snapshot)
DROP TABLE baseline_schema
SELECT * INTO baseline_schema FROM INFORMATION_SCHEMA.COLUMNS 
-- Type A: Test Pass
SELECT 
	CASE 
		WHEN count(*) = 0 THEN 'Test Pass'
		WHEN count(*) > 0 THEN 'Test Fail'
	END
	FROM (
		SELECT * FROM INFORMATION_SCHEMA.COLUMNS 
		EXCEPT
		SELECT * FROM baseline_schema
) as t

-- Type A: Test Fail
DELETE FROM baseline_schema WHERE COLUMN_NAME='name'
SELECT 
	CASE 
		WHEN count(*) = 0 THEN 'Test Pass'
		WHEN count(*) > 0 THEN 'Test Fail: ' + cast(count(*) as varchar) + ' records found. Expected 0'
	END
	FROM (
		SELECT * FROM INFORMATION_SCHEMA.COLUMNS 
		EXCEPT
		SELECT * FROM baseline_schema
) as t

-- Type B: Test Fail
IF EXISTS(
	SELECT * FROM INFORMATION_SCHEMA.COLUMNS 
	EXCEPT
	SELECT * FROM baseline_schema
	)
	RAISERROR ('Test Fail', 11, 1)
ELSE
	PRINT 'Test Pass'


----------
-- TEST CASE EXAMPLES
--
-- Practices wrapping SQL in test cases
----------
--BR1: All branches must have a manager
SELECT 
	CASE 
		WHEN count(*) = 0 THEN 'Test Pass'
		WHEN count(*) > 0 THEN 'Test Fail: ' + cast(count(*) as varchar) + ' records found. Expected 0'
	END
	FROM (
		SELECT * FROM Branches WHERE manager is null
	) as t
--or (show multiple rows - not prefered)
SELECT 
	CASE 
		WHEN count(*) = 0 THEN 'Test Pass'
		WHEN count(*) > 0 THEN 'Test Fail: ' + cast(count(*) as varchar) + ' records found. Expected 0'
	END, location
	FROM (
		SELECT * FROM Branches WHERE manager is null
	) as t
	GROUP BY location
--or (bad practice)
SELECT 'Test Fail', location FROM Branches WHERE manager is null
--or (test the data rather than the count - not prefered)
SELECT 
	CASE 
		WHEN location != 'Sydney' THEN 'Test Pass'
		WHEN location = 'Sydney' THEN 'Test Fail: ' + cast(count(*) as varchar) + ' records found. Expected 0'
	END, location
	FROM (
		SELECT * FROM Branches WHERE manager is null
	) as t
	GROUP BY location
--BR2: The account balance must = the sum of all transactions
IF EXISTS(
	SELECT sum(t.transaction_amt)-a.balance 
		FROM Transactions t, Accounts a
		WHERE t.account_id=a.account_id
		GROUP BY a.balance
		HAVING sum(t.transaction_amt)-a.balance > 0
	)
	RAISERROR ('Test Fail', 11, 1)
ELSE
	PRINT 'Test Pass'
--BR3: The manager must work at a branch
DECLARE @msg varchar(255)
DECLARE @result int
SET @result = 0
SET @msg = 'Test Fail: '
SET @result = (SELECT count(*) FROM (
	SELECT manager 
		FROM Branches b
		WHERE b.manager NOT IN (SELECT staff_id FROM WorksAt WHERE WorksAt.branch_id=b.name)
	) as t)
IF @result > 0
	BEGIN
		SET @msg = @msg + cast(@result as varchar) + ' records found. Expected 0'
		RAISERROR (@msg , 11, 100)
	END
ELSE
	PRINT 'Test Pass'
--BR4: The ages of each customer must be correct
SELECT 
	CASE 
		WHEN sumage = 0 THEN 'Test Pass'
		WHEN sumage > 0 THEN 'Test Fail'
	END
	FROM (
		SELECT SUM(datediff(yy, dob, getdate()) - coalesce(age, 0)) as sumage
			FROM Customers
	) as t
--BR5: All customers must have +’ve balances
IF EXISTS(
	SELECT balance FROM Accounts WHERE balance < 0
	)
	RAISERROR ('Test Fail', 11, 1)
ELSE
	PRINT 'Test Pass'
--or (via insert)
DECLARE @myrow INT
SET @myrow = 0
BEGIN TRANSACTION t1
	BEGIN TRY
		INSERT INTO Accounts VALUES (1, -100, 1)
		COMMIT TRANSACTION t1
	END TRY
	BEGIN CATCH
		ROLLBACK TRANSACTION t1
	END CATCH
SELECT 
	CASE 
		WHEN rownum = 0 THEN 'Test Pass'
		WHEN rownum > 0 THEN 'Test Fail: Negative balance was inserted'
	END
	FROM (
		SELECT @myrow as rownum
	) as t
--BR6: Managers must have the correct staff role
DECLARE @msg varchar(255)
DECLARE @result int
SET @result = 0
SET @msg = 'Test Fail: '
SET @result = (SELECT count(*) FROM (
	SELECT * 
		FROM staff s, branches b
		WHERE s.staff_id=b.manager AND s.role != 'Manager'
	) as t)
IF @result > 0
	BEGIN
		SET @msg = @msg + cast(@result as varchar) + ' records found. Expected 0'
		RAISERROR (@msg , 11, 100)
	END
ELSE
	PRINT 'Test Pass'
--BR7: All Customers must have an account
IF EXISTS(
	SELECT customer_id FROM Customers c
	EXCEPT
	SELECT customer_id FROM Accounts a
	)
	RAISERROR ('Test Fail', 11, 1)
ELSE
	PRINT 'Test Pass'
--BR8: All Staff must work at a Branch
DECLARE @msg varchar(255)
DECLARE @result int
SET @result = 0
SET @msg = 'Test Fail: '
SET @result = (SELECT count(*) FROM (
		SELECT staff_id FROM Staff
		EXCEPT 
		SELECT staff_id FROM WorksAt
	) as t)
IF @result > 0
	BEGIN
		SET @msg = @msg + cast(@result as varchar) + ' records found. Expected 0'
		RAISERROR (@msg , 11, 100)
	END
ELSE
	PRINT 'Test Pass'
--BR9: All Accounts must be opened by Staff
DECLARE @myrow INT
SET @myrow = 0
BEGIN TRANSACTION t1
	BEGIN TRY
		INSERT INTO Accounts VALUES (1, 100, NULL)
		COMMIT TRANSACTION t1
	END TRY
	BEGIN CATCH
		ROLLBACK TRANSACTION t1
	END CATCH
SELECT 
	CASE 
		WHEN rownum > 0 THEN 'Test Pass'
		WHEN rownum = 0 THEN 'Test Fail: Account was inserted without a staff member'
	END
	FROM (
		SELECT @myrow as rownum
	) as t


----------
-- ROW_NUMBER/RANK
--
-- Show how to return row numbers in the result set
----------
SELECT ROW_NUMBER() OVER (ORDER BY account_id), * FROM Transactions
SELECT ROW_NUMBER() OVER (ORDER BY transaction_amt), * FROM Transactions
SELECT ROW_NUMBER() OVER (PARTITION BY account_id ORDER BY transaction_amt), * FROM Transactions
SELECT RANK() OVER (ORDER BY account_id), * FROM Transactions
SELECT RANK() OVER (ORDER BY transaction_amt), * FROM Transactions
SELECT RANK() OVER (PARTITION BY account_id ORDER BY transaction_amt), * FROM Transactions

----------
-- SQL UPDATES, INSERTS & DELETE
--
-- Practice changing data in the database
----------
--Q1. Update Balance (Account)
INSERT INTO Transactions VALUES (1, 2000, getdate())
UPDATE Accounts 
	SET balance=(SELECT sum(t.transaction_amt) FROM transactions t WHERE Accounts.account_id=t.account_id)
SELECT * FROM accounts;
--Q2. Add 5 new Staff manually and 5 using SSMS
INSERT INTO Staff VALUES ('Tobias', 'Snoad', 'Teller')
INSERT INTO Staff VALUES ('Jennifer', 'Edmonson', 'Teller')
INSERT INTO Staff VALUES ('Ashley', 'Flynn', 'Teller')
INSERT INTO Staff VALUES ('Dev', 'Pathy', 'Teller')
INSERT INTO Staff VALUES ('Xianzheng', 'Zhou', 'Teller')
--Q3. Add a Manager for every Branch
UPDATE Branches SET manager=3 WHERE name=2;
UPDATE Branches SET manager=4 WHERE name=3;
--Q4. Update Age (Customer)
UPDATE Customers SET age=datediff(year, dob, getdate())
--Q5. Delete all Customers without an Account 
DELETE FROM Customers 
	WHERE customer_id NOT IN (SELECT customer_id FROM Accounts)
--Q6. Deposit $100 into Esthers Account
SELECT * FROM Customers c, Accounts a 
	WHERE a.customer_id=c.customer_id AND c.customer_id=1
INSERT INTO Transactions VALUES (1, 100, getdate());
UPDATE Accounts SET balance=balance+100 WHERE account_id=1
--or (via transaction)
BEGIN TRANSACTION t1
	BEGIN TRY
		INSERT INTO Transactions 
			VALUES (
				(SELECT TOP 1 a.account_id 
					FROM Customers c, Accounts a 
					WHERE a.customer_id=c.customer_id AND c.given_name='Esther'), 
				100, 
				getdate());

		UPDATE Accounts SET balance=balance+100 
			WHERE account_id=(
				SELECT TOP 1 a.account_id 
					FROM Customers c, Accounts a 
					WHERE a.customer_id=c.customer_id AND c.given_name='Esther')
		COMMIT TRANSACTION t1
	END TRY
	BEGIN CATCH
		ROLLBACK TRANSACTION t1
	END CATCH
--Q7. Convert all Customer given & family names to lower case
UPDATE Customers SET family_name=lower(family_name)
UPDATE Customers SET given_name=lower(given_name)
--Q8. Delete Customer #2
DELETE 
	FROM Transactions 
	WHERE account_id=(SELECT account_id FROM accounts where customer_id=2)
DELETE FROM Accounts WHERE customer_id=2
DELETE FROM Customers WHERE customer_id=2


----------
-- CREATING CUSTOMER DATA TYPES
--
-- For common constraints such as postcode or email address
----------
CREATE TYPE postcode FROM CHAR(4) NOT NULL;
CREATE TABLE foo (
	user_pc POSTCODE,
	username VARCHAR(20)
)
INSERT INTO foo VALUES ('1234', 'evan')
SELECT * FROM foo
INSERT INTO foo VALUES ('12345', 'evan')
SELECT * FROM foo
INSERT INTO foo VALUES ('x123', 'evan')
SELECT * FROM foo
DROP TABLE foo
DROP TYPE POSTCODE
-- or
CREATE TYPE postcode FROM INT 
	CHECK (postcode > 10000 and postcode < 99999);
CREATE TABLE foo (
	user_pc POSTCODE,
	username VARCHAR(20)
)
INSERT INTO foo VALUES (11111, 'evan')
SELECT * FROM foo
INSERT INTO foo VALUES (123456, 'evan')
SELECT * FROM foo
INSERT INTO foo VALUES (100, 'evan')
SELECT * FROM foo


----------
-- DDL
--
-- Changing the database
----------
-- ALTER TABLE
ALTER TABLE Customers ADD salutation VARCHAR(10);
ALTER TABLE Customers ALTER COLUMN salutation varchar(5);
ALTER TABLE Customers ALTER COLUMN salutation varchar(5) NOT NULL;
ALTER TABLE Customers 
	ADD CONSTRAINT cons_name DEFAULT 'Mr' FOR salutation ;
ALTER TABLE Customers DROP COLUMN salutation;
-- ALTER TABLE: Computed Columns
ALTER TABLE Customers 
	ADD fullname2 AS given_name + ' ' + family_name 
ALTER TABLE Customers 
	ADD fullname AS given_name + ' ' + family_name PERSISTED
SELECT * FROM Customers;
-- CREATE TABLE
CREATE TABLE Address (
	addressid INT IDENTITY (1,1),
	street VARCHAR(50),
	suburb VARCHAR(50),
	city VARCHAR(50),
	country VARCHAR(50) DEFAULT 'Malaysia',
	postcode INT CHECK (postcode > 1000 AND postcode < 9999)
)
INSERT INTO Address VALUES (NULL, NULL, NULL, DEFAULT, 1001)
SELECT * FROM Address
-- DROP TABLE
DROP TABLE Customer
-- CREATE INDEX
CREATE INDEX index_name ON Customer (family_name);
CREATE INDEX index_name2 ON Customer (family_name, given_name);
-- CREATE VIEW
CREATE VIEW FemaleCustomers AS SELECT * from Customer 
	WHERE gender='F'; 
SELECT * FROM FemaleCustomers;
DROP VIEW FemaleCustomers;

